import os
from os.path import join, expanduser
import subprocess as sp
import shutil
import http.server
import threading
from time import sleep

from fire import Fire
from dotenv import load_dotenv
import shlex

import atexit

from .ssdp import discover

load_dotenv(join(os.getcwd(), '.env'))
USER_HOME = expanduser("~")


def get_env_or_opt(opt, env_name):
    return opt or os.getenv(env_name)


class WinCast:

    @property
    def ffmpeg_home(self):
        return os.getenv('FFMPEG_HOME')

    @property
    def ffmpeg_bin(self):
        default_bin = 'ffmpeg.exe' if self.ffmpeg_home is None else join(
            self.ffmpeg_home, 'ffmpeg.exe')
        return os.getenv('FFMPEG_BIN', default_bin)

    @property
    def dlna_cast_dir(self):
        return os.getenv('DLNA_CAST_DIR', join(USER_HOME, 'dlna-cast'))

    def _get_devices(self):
        devices = discover(timeout=10)
        return [d for d in devices if d.find_action('SetAVTransportURI')]

    def _find_device(self, name):
        for d in self._get_devices():
            if d.friendly_name == name:
                return d

    def _start_http_server(self):
        class Handler(http.server.SimpleHTTPRequestHandler):
            def __init__(_, *args, **kwargs):
                super().__init__(*args, directory=self.dlna_cast_dir, **kwargs)

        with http.server.ThreadingHTTPServer(("", 0), Handler) as httpd:
            self._httpd = httpd
            self._listen_port = httpd.server_address[1]
            print("HTTP: start to serving at port", self._listen_port)
            httpd.serve_forever()

    def _start_ffmpeg_streaming(self, framerate=30, input_opts='', enc_opts=''):
        input_opts = input_opts or '-f dshow -i video="screen-capture-recorder":audio="virtual-audio-capturer"'
        enc_opts = enc_opts or '-c:v libx264 -preset fast -tune zerolatency -crf 21 -vf format=yuv420p'
        cmd = [
            self.ffmpeg_bin, '-framerate', str(framerate),
            input_opts,
            enc_opts,
            '-f hls -hls_time 3 -hls_list_size 10 -hls_flags delete_segments',
            join(self.dlna_cast_dir, 'index.m3u8'),
        ]
        cmd = ' '.join(cmd)
        print('run command: ', cmd)
        self._ffmpeg_process = sp.Popen(cmd)
        atexit.register(lambda: self._ffmpeg_process.kill())

    def __init__(self):
        self._listen_port = None
        self._httpd = None
        self._ffmpeg_process = None

    def screen(self, dlna_device=None, framerate=30, input_opts=None, enc_opts=None):
        # start ffmpeg streaming
        input_opts = get_env_or_opt(input_opts, 'FFMPEG_INPUT_OPTS')
        enc_opts = get_env_or_opt(input_opts, 'FFMPEG_ENC_OPTS')

        shutil.rmtree(self.dlna_cast_dir, ignore_errors=True)
        os.makedirs(self.dlna_cast_dir, exist_ok=True)
        self._start_ffmpeg_streaming(framerate, input_opts, enc_opts)

        # discover dlan_device
        dlna_device = get_env_or_opt(dlna_device, 'DLNA_DEVICE')
        assert dlna_device, 'dlna_device must be set!'
        device = self._find_device(dlna_device)
        assert device, 'cannot find deivce named {}'.format(dlna_device)

        # start http server for hls
        thread = threading.Thread(target=self._start_http_server, daemon=True)
        thread.start()

        while self._listen_port is None:
            sleep(1)  # TODO: use thread.Event instead

        # play video
        hls_url = 'http://{}:{}/index.m3u8'.format(
            device.iface_ip, self._listen_port)
        print('start to play {} on {}'.format(hls_url, device.friendly_name))
        device.AVTransport.SetAVTransportURI(
            InstanceID=0,
            CurrentURI=hls_url,
            CurrentURIMetaData=''
        )

        def stop_play_on_exit():
            device.AVTransport.Stop(
                InstanceID=0,
            )
            print('stop remote video')

        atexit.register(stop_play_on_exit)

        # wait for subprocess
        self._ffmpeg_process.wait()

    def list_dshow_devices(self):
        cmd = [self.ffmpeg_bin, '-hide_banner',
               '-list_devices', 'true', '-f', 'dshow', '-i', 'dummy']
        sp.call(cmd)

    def list_dlna_devices(self):
        devices = self._get_devices()
        for d in devices:
            print(d.friendly_name)

def main():
    Fire(WinCast)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dlna_cast']

package_data = \
{'': ['*']}

install_requires = \
['fire>=0.4.0,<0.5.0',
 'python-dotenv[cli]>=0.21.0,<0.22.0',
 'upnpclient>=1.0.3,<2.0.0']

entry_points = \
{'console_scripts': ['dlna-cast = dlna_cast.main:main']}

setup_kwargs = {
    'name': 'dlna-cast',
    'version': '0.1.4',
    'description': 'A cross-platform command-line tool that casts screen and media file to remote DLNA device.',
    'long_description': '# dlna-cast\nA cross-platform command-line tool that casts screen and media file to remote DLNA device.\n\n## Introduction\n`dlna-cast` uses `ffmpeg` to capture screen and audio, then convert them into HLS streams which could be served by a simple HTTP server. The HLS url will be send to the selected device via uPnP protocol and then you can watch you screen on the remote device (smart TV, typically).\n\nThis tool is supposed to be cross-platform but currently I don\'t have a Linux or MacOS device at hand so it can only run on Windows now. It won\'t be hard to support other platforms though, as there are no platform specific dependencies.\n\nHLS is chosen just because it is easy to implement. But the problem of HLS is its high latency (up to 5-10s or more) so it\'s definitely not for scenarios that require low latency (presentation for example). But as a trade-off the streaming quality exceeds a lot of software-based screen-casting solutions (Lebocast for example) that have been tested by myself, which make it pretty good to stream music or video playing on your PC to the supported TV.\n\n## Install\n```bash\npip install dlna-cast\n```\nPlease ensure your Python is 3.7 or above.\n\n### Install ffmpeg\nYou can install `ffmpeg` by compiling from source code, or just download the prebuild binary from https://ffmpeg.org/download.html\n\nYou need to ensure the `ffmpeg` command can be found in the `PATH` environment variable, or else you need to set `FFMPEG_BIN` or `FFMPEG_HOME` to let `dlna-cast` know where to find the command. \n\n`dlna-cast` supports reading the environment variable from `.env` file.  You can create a `.env` file under the folder you are gonna run the `dlna-cast` command with the following content.\n\n```bash\nFFMPEG_BIN=D:\\ffmpeg\\ffmpeg.exe\n# or\nFFMPEG_HOME=D:\\ffmpeg\n```\n\nYou can also use the `dotenv set` command to update the `.env` file and use the `dotenv list` to check result.\n\n```bash\ndotenv set FFMPEG_HOME "D:\\ffmpeg"\ndotenv list\n```\n\n### Install ScreenCapturerRecorder on Windows\nThough `ffmpeg` is shipped with `gdigrab` to capture screens on Windows, its performance is terrible when frame rate is high. `dlna-cast` uses ScreenCapturerRecorder for the sake of performance. You need to [download](https://github.com/rdp/screen-capture-recorder-to-video-windows-free/releases) and install it before starting to use this tool.\n\n## Get Started\nBefore you start to stream your screen to remote devices that support DLNA protocol, you need to discover available devices in your LAN by running the following command.\n\n```bash\ndlna-cast list_dlna_devices\n# You will see the output if supported devices are found\nHuaweiPro\nLebocast\n```  \n\nAnd now you can cast your screen to one of the found devices by running the following command.\n```bash\ndlna-cast screen --dlna_device HuaweiPro\n``` \n\nOr you can also set `DLNA_DEVICE` in the `.env` file so that you can skip to set `--dlna_device` next time.\n\n```bash\ndotenv set DLNA_DEVICE HuaweiPro\ndlna-cast screen\n```\n\nYou can stop casting by pressing `Ctrl+C`. \n',
    'author': 'weihong.xu',
    'author_email': 'xuweihong.cn@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
